"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.syncErrorHandler = exports.Synchronizer = exports.Contracts = exports.Chain = void 0;
// External modules
const events_1 = __importDefault(require("events"));
const crypto_1 = require("crypto");
const ethers_1 = require("ethers");
// Our local types
const deth_1 = require("./deth");
// Local modules
const sdk_data_1 = require("@tornado/sdk-data");
const sdk_utils_1 = require("@tornado/sdk-utils");
// @ts-ignore
const pouchdb_collate_1 = require("pouchdb-collate");
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ REST ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/**
 * The Chain class stores Tornado-agnostic chain data and also
 * handles such interactions.
 */
class Chain {
    provider;
    _emptySigner;
    _signer;
    _chainId;
    _symbol;
    _fetched;
    constructor(provider, signer) {
        this.provider = provider;
        this._signer = signer;
        this._emptySigner = new ethers_1.VoidSigner('0x' + (0, crypto_1.randomBytes)(20).toString('hex'), provider);
        this._fetched = false;
    }
    connectSigner(signer) {
        this._signer = signer;
    }
    async fetchChainData() {
        const network = await this.provider.getNetwork();
        this._chainId = network.chainId;
        this._symbol = await sdk_data_1.Onchain.getNetworkSymbol(String(network.chainId));
        this._fetched = true;
    }
    _signerConnected(parentCallName) {
        if (!this._signer)
            throw sdk_utils_1.ErrorUtils.getError(`Chain.${parentCallName}: signer must be connected!`);
    }
    _propertiesFetched(parentCallName) {
        if (!this._fetched)
            throw sdk_utils_1.ErrorUtils.getError(`Chain.${parentCallName}: properties must be fetched first with \`fetchProperties\`.`);
    }
    get signer() {
        this._signerConnected('signer');
        return this._signer;
    }
    get id() {
        this._propertiesFetched('id');
        return this._chainId;
    }
    get symbol() {
        this._propertiesFetched('symbol');
        return this._symbol;
    }
    latestBlockNum() {
        return this.provider.getBlockNumber();
    }
    getAccountBalance(account) {
        return this.provider.getBalance(account);
    }
    getGasPrice() {
        return this.provider.getGasPrice();
    }
    getTokenContract(tokenAddress) {
        return Contracts.getToken(tokenAddress, this._signer ?? this.provider);
    }
    async getTokenDecimals(token) {
        let treq = {
            to: token,
            data: '0x313ce567'
        };
        return ethers_1.BigNumber.from(await this._emptySigner.call(treq));
    }
    async getTokenBalance(account, token, normalized = false) {
        let treq = {
            to: token,
            data: '0x70a08231000000000000000000000000' + sdk_utils_1.HexUtils.prepareAddress(account)
        };
        let divisor = normalized ? ethers_1.BigNumber.from(10).pow(await this.getTokenDecimals(token)) : 1;
        return ethers_1.BigNumber.from(await this._emptySigner.call(treq)).div(divisor);
    }
    async populateBatchCall(callStruct) {
        if (callStruct[0].value)
            return await deth_1.Multicall3Contract__factory.connect(await sdk_data_1.Onchain.getMulticall3Address(String(this.id)), this.provider).populateTransaction.aggregate3Value(callStruct);
        return await deth_1.Multicall3Contract__factory.connect(await sdk_data_1.Onchain.getMulticall3Address(String(this.id)), this.provider).populateTransaction.aggregate3(callStruct);
    }
    async batchCall(callStruct) {
        if (this._signer)
            if (callStruct[0].value)
                return await deth_1.Multicall3Contract__factory.connect(await sdk_data_1.Onchain.getMulticall3Address(String(this.id)), this._signer).aggregate3Value(callStruct);
            else {
                return await deth_1.Multicall3Contract__factory.connect(await sdk_data_1.Onchain.getMulticall3Address(String(this.id)), this.provider).aggregate3(callStruct);
            }
        else
            throw sdk_utils_1.ErrorUtils.getError('Chain.batchCall: no signer provided.');
    }
}
exports.Chain = Chain;
/**
 * This is Tornado-specific.
 */
var Contracts;
(function (Contracts) {
    function _getContract(name, address, signerOrProvider) {
        if (name == 'TornadoInstance') {
            return deth_1.TornadoInstance__factory.connect(address, signerOrProvider);
        }
        else if (name == 'TornadoProxy') {
            return deth_1.TornadoProxy__factory.connect(address, signerOrProvider);
        }
        else if (name == 'ETHTornado') {
            return deth_1.ETHTornado__factory.connect(address, signerOrProvider);
        }
        else if (name == 'ERC20Tornado') {
            return deth_1.ERC20Tornado__factory.connect(address, signerOrProvider);
        }
        else if (name == 'RelayerRegistry') {
            return deth_1.RelayerRegistry__factory.connect(address, signerOrProvider);
        }
        else {
            return deth_1.ERC20__factory.connect(address, signerOrProvider);
        }
    }
    const contractMap = new Map();
    function getProxy(network, signerOrProvider) {
        const key = `TornadoProxy${network}`;
        if (!contractMap.has(key)) {
            contractMap.set(key, _getContract('TornadoProxy', sdk_data_1.Onchain.getProxyAddressSync(network), signerOrProvider));
        }
        return contractMap.get(`TornadoProxy${network}`);
    }
    Contracts.getProxy = getProxy;
    function getRegistry(signerOrProvider, network) {
        const key = `RelayerRegistry${network}`;
        if (!contractMap.has(key))
            contractMap.set(key, _getContract('RelayerRegistry', sdk_data_1.Onchain.getRegistryAddressSync(network), signerOrProvider));
        return contractMap.get(key);
    }
    Contracts.getRegistry = getRegistry;
    function getInstance(network, token, denomination, signerOrProvider) {
        const key = `TornadoInstance${network}${token}${denomination}`;
        if (!contractMap.has(key)) {
            contractMap.set(key, _getContract('TornadoInstance', sdk_data_1.Onchain.getInstanceAddressSync(network, token, denomination), signerOrProvider));
        }
        return contractMap.get(key);
    }
    Contracts.getInstance = getInstance;
    function getToken(tokenAddress, signerOrProvider) {
        if (!contractMap.has(tokenAddress))
            contractMap.set(tokenAddress, _getContract('ERC20', tokenAddress, signerOrProvider));
        return contractMap.get(tokenAddress);
    }
    Contracts.getToken = getToken;
    function getTornToken(signerOrProvider) {
        const key = '$TORN';
        if (!contractMap.has(key)) {
            contractMap.set(key, _getContract('ERC20', '0x77777feddddffc19ff86db637967013e6c6a116c', signerOrProvider));
        }
        return contractMap.get(key);
    }
    Contracts.getTornToken = getTornToken;
})(Contracts = exports.Contracts || (exports.Contracts = {}));
class Synchronizer extends events_1.default {
    caches;
    constructor() {
        super();
        this.caches = new Map();
    }
    async sync(eventName, filter, contract, cache, options) {
        const _options = await this._populateSyncOptions(options ?? {});
        // Assign pooler
        cache.initializePooler(cache.getCallbacks(contract), cache.getErrorHandlers(), _options.concurrencyLimit);
        // Decide whether we have a latest block
        const numEntries = (await cache.db.info()).doc_count;
        // Check for synced blocks
        if (0 < numEntries) {
            const [lastSyncedBlock, ,] = (0, pouchdb_collate_1.parseIndexableString)((await cache.db.allDocs({ descending: true, limit: 1 })).rows[0].id);
            _options.startBlock = lastSyncedBlock < _options.startBlock ? _options.startBlock : lastSyncedBlock;
            _options.blockDelta = _options.blockDivisor
                ? Math.floor((_options.targetBlock - _options.startBlock) / _options.blockDivisor)
                : _options.blockDelta;
            if (10000 < _options.blockDelta)
                throw sdk_utils_1.ErrorUtils.getError('Synchronizer.sync: blockDelta must not be above 10,000!');
        }
        // Start synchronizing
        let dbPromises = [];
        this.emit('debug', _options.startBlock, _options.targetBlock, _options.blockDelta);
        this.emit('sync', 'syncing');
        for (let currentBlock = _options.startBlock, blockDelta = _options.blockDelta, targetBlock = _options.targetBlock, concurrencyLimit = _options.concurrencyLimit; currentBlock < targetBlock; currentBlock += blockDelta) {
            if (cache.pooler.pending < concurrencyLimit) {
                const sum = currentBlock + blockDelta;
                await sdk_utils_1.AsyncUtils.timeout(_options.msTimeout);
                if (currentBlock + blockDelta < targetBlock) {
                    await cache.pooler.pool(currentBlock, sum);
                }
                else {
                    await cache.pooler.pool(currentBlock, sum - (sum % targetBlock));
                }
                this.emit('debug', currentBlock++, sum);
            }
            else {
                let res = await cache.pooler.race();
                if (res.length != 0)
                    dbPromises.push(cache.db.bulkDocs(res.map((el) => cache.buildDoc(el))).catch((err) => {
                        throw sdk_utils_1.ErrorUtils.ensureError(err);
                    }));
                currentBlock -= blockDelta;
            }
        }
        this.emit('sync', 'synced');
        // Immediately start listening if we're doing this
        if (_options.listenForEvents)
            this.listenForEvents(eventName, contract, filter, cache);
        // Then wait for all pooler requests to resolve
        let results = await cache.pooler.all();
        // Then transform them, we know the shape in forward
        results = results.reduce((res, response) => {
            if (response[0])
                response.forEach((el) => res.push(cache.buildDoc(el)));
            return res;
        }, []);
        // Then wait for old dbPromises to resolve
        await Promise.all(dbPromises);
        // Add the last docs
        await cache.db.bulkDocs(results).catch((err) => {
            throw sdk_utils_1.ErrorUtils.ensureError(err);
        });
        // Add cache if not present
        if (!this.caches.has(cache.name))
            this.caches.set(cache.name, cache);
    }
    loadCache(name) {
        return this.caches.get(name);
    }
    listenForEvents(name, contract, filter, cache) {
        contract.on(filter, (...eventArgs) => {
            this.emit(name, cache.name, cache.db.put(cache.buildDoc(eventArgs[eventArgs.length - 1])));
        });
    }
    clearListenerByIndex(contract, event, listenerIndex = 0) {
        const listeners = contract.listeners();
        contract.off(event, listeners[listenerIndex]);
    }
    async _populateSyncOptions(options) {
        if (!options.targetBlock)
            throw sdk_utils_1.ErrorUtils.getError('Synchronizer._populateSyncOptions: targetBlock not set.');
        if (!options.startBlock)
            throw sdk_utils_1.ErrorUtils.getError('Synchronizer._populateSyncOptions: startBlock not set.');
        options.blockDelta =
            options.blockDelta ??
                (options.blockDivisor
                    ? Math.floor((options.targetBlock - options.startBlock) / options.blockDivisor)
                    : 10000);
        if (10000 < options.blockDelta)
            throw sdk_utils_1.ErrorUtils.getError('Synchronizer._populateSyncOptions: blockDelta must not be above 10,000!');
        options.concurrencyLimit = options.concurrencyLimit ?? 8;
        options.msTimeout = options.msTimeout ?? 200; // 5 requests per second
        options.listenForEvents = options.listenForEvents ?? false;
        return options;
    }
}
exports.Synchronizer = Synchronizer;
function syncErrorHandler(err, numResolvedPromises, callbackIndex, orderIndex, ...args) {
    err = sdk_utils_1.ErrorUtils.ensureError(err);
    if (err.message.match('context deadline exceeded'))
        console.error(sdk_utils_1.ErrorUtils.getError(`Context deadline exceeded, stop if more promises do not resolve. Resolved: ${numResolvedPromises}`));
    else if (err.message.match('Invalid JSON RPC'))
        console.error(sdk_utils_1.ErrorUtils.getError('Endpoint returned invalid value (we might be rate limited), retrying.'));
    else if (err.message.match('logs.forEach'))
        console.error(sdk_utils_1.ErrorUtils.getError("Unclear type mismatch error, retrying. If this doesn't stop, cancel."));
    else {
        err.message += `\nCallback args supplied: [${args.join(', ')}]\n`;
        throw err;
    }
}
exports.syncErrorHandler = syncErrorHandler;
//# sourceMappingURL=index.js.map